import datetime
from typing import Union, Optional, Iterable
from pathlib import Path

import pandas as pd

from reki.data_finder._config import (
    find_config, load_config, get_default_local_config_path,
)
from reki.data_finder._util import find_file, find_files, render_file_name


def find_local_file(
        data_type: str,
        start_time: Union[str, pd.Timestamp, datetime.datetime],
        forecast_time: Union[str, pd.Timedelta] = "0",
        data_level: Optional[Union[str, Iterable]] = ("archive", "storage"),
        path_type: str = "local",
        data_class: str = "od",
        config_dir: Union[str, Path] = None,
        obs_time: Union[str, pd.Timestamp] = None,
        debug: bool = False,
        **kwargs,
) -> Optional[Path]:
    """
    Find local data path using config files in config dir.

    Parameters
    ----------
    data_type
        data type, relative path of config file to `config_dir` without suffix.
        For example `grapes_gfs_gmf/grib2/orig` means using config file `{config_dir}/grapes_gfs_gmf/grib2/orig.yaml`.
    start_time
        start time of production. YYYYMMDDHH if str.
    forecast_time
        forecast time of production. A string (such as `3h`) will be parsed by ``pd.to_timedelta``.
    data_level
        data storage level, ["archive", "runtime", "storage", ... ], default is ``("archive", "storage")``.
    path_type
        path type, ["local", "storage", ...], for future usage.
    data_class
        data class, ``od`` means operation systems.
    config_dir
        config root directory. If None, use embedded config files in `conf` directory.
    obs_time
        time for observation data.
    debug
        show debug info.
    **kwargs
        other options needed by path template. All of them will be added into `query_vars`.

    Returns
    -------
    Path or None
        file path if found or None if not.

    Examples
    --------
    Find an existing orig grib2 file of CMA-GFS in CMA-PAI HPC.

    >>> find_local_file(
    ...     "cma_gfs_gmf/grib2/orig",
    ...     start_time="2023122000",
    ...     forecast_time="3h",
    ... )
    PosixPath('/g1/COMMONDATA/OPER/NWPC/GRAPES_GFS_GMF/Prod-grib/2023122000/ORIG/gmf.gra.2023122000003.grb2')

    Find a non-existing orig grib2 file of CMA-GFS.

    >>> find_local_file(
    ...     "cma_gfs_gmf/grib2/orig",
    ...     start_time="2020032100",
    ...     forecast_time="1h",
    ... )
    None

    Find a grib2 file in storage for CMA-MESO 3km.

    >>> find_local_file(
    ...     "cma_meso_3km/grib2/orig",
    ...     start_time="2020032100",
    ...     forecast_time="1h",
    ...     data_level="storage",
    ... )
    PosixPath('/sstorage1/COMMONDATA/OPER/NWPC/GRAPES_MESO_3KM/Prod-grib/2020032100/ORIG/rmf.hgra.2020032100001.grb2')

    Find a grib2 file in storage for CMA-GEPS.

    >>> find_local_file(
    ...     "cma_geps/grib2/orig",
    ...     start_time="2023032100",
    ...     forecast_time="3h",
    ...     data_level="storage",
    ...     number=1,
    ... )
    PosixPath('/sstorage1/COMMONDATA/OPER/NWPC/GRAPES_GEPS/Prod-grib/2023032100/grib2/gef.gra.001.2023032100003.grb2')

    Find postvar ctl file for CMA-TYM in Windows mount storage.

    .. code-block:: pycon

        >>> find_local_file(
        ...     "cma_tym/bin/postvar_ctl",
        ...     start_time="2021080200",
        ...     storage_base="M:"
        ... )
        WindowsPath('M:/GRAPES_TYM/Fcst-main/2021080200/post.ctl_2021080200')

    """
    if config_dir is None:
        config_dir = get_default_local_config_path()

    config_file_path = find_config(config_dir, data_type, data_class)
    if config_file_path is None:
        raise ValueError(f"data type is not found: {data_type}")

    if isinstance(forecast_time, str):
        forecast_time = pd.to_timedelta(forecast_time)
    if isinstance(start_time, str):
        start_time = pd.to_datetime(start_time, format="%Y%m%d%H")
    if isinstance(obs_time, str):
        obs_time = pd.to_datetime(obs_time)
    elif obs_time is None:
        obs_time = start_time

    config = load_config(config_file_path)
    file_path = find_file(
        config,
        data_level,
        start_time,
        forecast_time,
        obs_time=obs_time,
        debug=debug,
        **kwargs
    )
    return file_path


def get_local_file_name(
        data_type: str,
        start_time: Union[str, pd.Timestamp, datetime.datetime],
        forecast_time: Union[str, pd.Timedelta] = pd.Timedelta(hours=1),
        data_class: str = "od",
        config_dir: Union[str, Path] = None,
        obs_time: Union[str, pd.Timestamp] = None,
        **kwargs,
) -> str:
    if config_dir is None:
        config_dir = get_default_local_config_path()

    config_file_path = find_config(config_dir, data_type, data_class)
    if config_file_path is None:
        raise ValueError(f"data type is not found: {data_type}")

    # if isinstance(forecast_time, str):
    #     forecast_time = pd.to_timedelta(forecast_time)

    if isinstance(start_time, str):
        start_time = pd.to_datetime(start_time, format="%Y%m%d%H")
    if isinstance(obs_time, str):
        obs_time = pd.to_datetime(obs_time)
    elif obs_time is None:
        obs_time = start_time

    config = load_config(config_file_path)
    file_name = render_file_name(
        config,
        start_time,
        forecast_time,
        obs_time=obs_time,
        **kwargs
    )
    return file_name


def find_local_files(
        data_type: str,
        start_time: Union[str, pd.Timestamp, datetime.datetime],
        forecast_time: Union[str, pd.Timedelta] = "0",
        data_level: Optional[Union[str, Iterable]] = ("archive", "storage"),
        path_type: str = "local",
        data_class: str = "od",
        config_dir: Union[str, Path] = None,
        glob: bool = True,
        **kwargs,
) -> Optional[Path]:
    if config_dir is None:
        config_dir = get_default_local_config_path()

    config_file_path = find_config(config_dir, data_type, data_class)
    if config_file_path is None:
        raise ValueError(f"data type is not found: {data_type}")

    if isinstance(forecast_time, str):
        forecast_time = pd.to_timedelta(forecast_time)
    if isinstance(start_time, str):
        start_time = pd.to_datetime(start_time, format="%Y%m%d%H")

    config = load_config(config_file_path)
    file_path = find_files(config, data_level, start_time, forecast_time, glob, **kwargs)
    return file_path



from dataclasses import dataclass, asdict
from typing import List, Dict, Union, Optional

import pytest

from reki.format.grib.eccodes import load_field_from_file


@dataclass
class QueryOption:
    parameter: str
    level_type: Union[str, Dict]
    level: Optional[Union[float, Dict]]
    level_dim: Optional[str] = None


@dataclass
class TestCase:
    query: QueryOption
    expected_level_name: Optional[str]
    expected_level: float


def test_pl(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=1.5, level_dim=None),
            expected_level_name="pl",
            expected_level=1.5
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=1.5, level_dim="pl"),
            expected_level_name="pl",
            expected_level=1.5
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=1.5, level_dim="isobaricInhPa"),
            expected_level_name="isobaricInhPa",
            expected_level=1.5
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=1.5, level_dim="isobaricInPa"),
            expected_level_name="isobaricInPa",
            expected_level=150.
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=1.5, level_dim="other_level_string"),
            expected_level_name="other_level_string",
            expected_level=1,
        )  # TODO: use float point?
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.coords[test_case.expected_level_name].values == test_case.expected_level


def test_ml(modelvar_file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="gh", level_type="ml", level=10, level_dim=None),
            expected_level_name="ml",
            expected_level=10
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type="ml", level=10, level_dim="ml"),
            expected_level_name="ml",
            expected_level=10,
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type="ml", level=10, level_dim="other_level_string"),
            expected_level_name="other_level_string",
            expected_level=10,
        ),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            modelvar_file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.coords[test_case.expected_level_name].values == test_case.expected_level


def test_sfc(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type="sfc", level=None, level_dim=None),
            expected_level_name="sfc",
            expected_level=0,
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="sfc", level=None, level_dim="sfc"),
            expected_level_name="sfc",
            expected_level=0
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="sfc", level=None, level_dim="other_level_string"),
            expected_level_name="other_level_string",
            expected_level=0
        ),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.coords[test_case.expected_level_name].values == test_case.expected_level


def test_type_of_level(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="gh", level_type="isobaricInhPa", level=10, level_dim=None),
            expected_level_name="isobaricInhPa",
            expected_level=10
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type="isobaricInhPa", level=10, level_dim="other_level_string"),
            expected_level_name="other_level_string",
            expected_level=10
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type="isobaricInPa", level=50, level_dim=None),
            expected_level_name="isobaricInPa",
            expected_level=50
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type="isobaricInPa", level=50, level_dim="other_level_string"),
            expected_level_name="other_level_string",
            expected_level=50
        ),
        TestCase(
            query=QueryOption(parameter="ulwrf", level_type="nominalTop", level=None, level_dim=None),
            expected_level_name="nominalTop",
            expected_level=0
        ),
        TestCase(
            query=QueryOption(parameter="tmax", level_type="heightAboveGround", level=2, level_dim=None),
            expected_level_name="heightAboveGround",
            expected_level=2
        ),
        TestCase(
            query=QueryOption(parameter="prmsl", level_type="meanSea", level=None, level_dim=None),
            expected_level_name="meanSea",
            expected_level=0
        ),
        TestCase(
            query=QueryOption(parameter="q", level_type="depthBelowLandLayer", level={"first_level": 0.1, "second_level": 0.4}, level_dim=None),
            expected_level_name="depthBelowLandLayer",
            expected_level=0,
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="depthBelowLandLayer", level={"first_level": 1, "second_level": 2}, level_dim=None),
            expected_level_name="depthBelowLandLayer",
            expected_level=1,
        ),
        TestCase(
            query=QueryOption(parameter="vwsh", level_type="heightAboveGroundLayer", level={"first_level": 1000, "second_level": 0}, level_dim=None),
            expected_level_name="heightAboveGroundLayer",
            expected_level=1000,
        ),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query),
        )
        assert field is not None
        assert field.coords[test_case.expected_level_name].values == test_case.expected_level


def test_grib_key_for_pl(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level=10, level_dim=None),
            expected_level_name="isobaricInhPa",
            expected_level=10
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level=10, level_dim="pl"),
            expected_level_name="pl",
            expected_level=10
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level=10, level_dim="isobaricInhPa"),
            expected_level_name="isobaricInhPa",
            expected_level=10
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level=10, level_dim="isobaricInPa"),
            expected_level_name="isobaricInPa",
            expected_level=1000
        ),

        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 150}, level_dim=None),
            expected_level_name="isobaricInhPa",
            expected_level=1,
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 150}, level_dim="pl"),
            expected_level_name="pl",
            expected_level=1.5,
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 150}, level_dim="isobaricInhPa"),
            expected_level_name="isobaricInhPa",
            expected_level=1.5,
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 150}, level_dim="isobaricInPa"),
            expected_level_name="isobaricInPa",
            expected_level=150,
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 150}, level_dim="other_level_string"),
            expected_level_name="other_level_string",
            expected_level=1,
        ),

        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 50}, level_dim=None),
            expected_level_name="isobaricInPa",
            expected_level=50
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 50}, level_dim="pl"),
            expected_level_name="pl",
            expected_level=0.5
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 50}, level_dim="isobaricInhPa"),
            expected_level_name="isobaricInhPa",
            expected_level=0.5
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 50}, level_dim="isobaricInPa"),
            expected_level_name="isobaricInPa",
            expected_level=50
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type={"typeOfFirstFixedSurface": "pl"}, level={"first_level": 50}, level_dim="other_level_string"),
            expected_level_name="other_level_string",
            expected_level=50
        ),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.coords[test_case.expected_level_name].values == test_case.expected_level


def test_grib_key_modelvar(modelvar_file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type={"typeOfFirstFixedSurface": 131}, level=10, level_dim=None),
            expected_level_name="level_131",
            expected_level=10,
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type={"typeOfFirstFixedSurface": 131}, level=10, level_dim="ml"),
            expected_level_name="ml",
            expected_level=10,
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type={"typeOfFirstFixedSurface": 131}, level=10, level_dim="other_level_string"),
            expected_level_name="other_level_string",
            expected_level=10,
        ),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            modelvar_file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.coords[test_case.expected_level_name].values == test_case.expected_level



from dataclasses import dataclass, asdict
from typing import Union, Dict, Optional, List

from reki.format.grib.eccodes import load_field_from_file


@dataclass
class QueryOption:
    parameter: Union[str, dict]
    level_type: Union[str, Dict]
    level: Optional[Union[float, Dict]]


@dataclass
class TestCase:
    query: QueryOption


def test_short_name(file_path):
    test_cases = [
        TestCase(query=QueryOption(parameter="t", level_type="pl", level=850))
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None


def test_numbers(file_path):
    """
    雷达组合反射率
    """
    test_cases = [
        TestCase(
            query=QueryOption(parameter={
                "discipline": 0,
                "parameterCategory": 16,
                "parameterNumber": 225,
            }, level_type="pl", level=850)
        )
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None


def test_embedded_short_name(file_path):
    test_cases = [
        TestCase(query=QueryOption(parameter="DEPR", level_type="pl", level=850)),
        TestCase(query=QueryOption(parameter={
            "discipline": 0,
            "parameterCategory": 0,
            "parameterNumber": 7,
        }, level_type="pl", level=850)),
    ]

    fields = []

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        fields.append(field)

    assert fields[0].attrs["GRIB_count"] == fields[1].attrs["GRIB_count"]



import abc
from typing import Union, Literal

import numpy as np
import xarray as xr


class BaseInterpolator(abc.ABC):
    @abc.abstractmethod
    def interpolate_grid(
            self,
            data: xr.DataArray,
            target: xr.DataArray,
    ) -> xr.DataArray:
        pass

    @abc.abstractmethod
    def extract_point(
            self,
            data: xr.DataArray,
            latitude: Union[int, float],
            longitude: Union[int, float],
    ) -> xr.DataArray:
        pass


def _get_interpolator(
        scheme: str,
        engine: Literal["scipy", "xarray"],
        **kwargs
) -> BaseInterpolator:
    if engine == "scipy":
        from ._scipy import ScipyInterpnInterpolator, ScipyRectBivariateSplineInterpolator
        if scheme in ["linear", "nearest", "splinef2d"]:
            return ScipyInterpnInterpolator(scheme, **kwargs)
        elif scheme in ["rect_bivariate_spline"]:
            return ScipyRectBivariateSplineInterpolator(scheme, **kwargs)
        else:
            raise ValueError(f"{scheme} is not supported for engine {engine}")
    elif engine == "xarray":
        from ._xarray import XarrayInterpolator
        return XarrayInterpolator(scheme, **kwargs)
    else:
        raise ValueError(f"engine {engine} is not supported")


def _create_data_array(
        data: xr.DataArray,
        target: xr.DataArray,
        target_values: np.ndarray,
) -> xr.DataArray:
    coords = data.coords.to_dataset()
    coords["longitude"] = target["longitude"]
    coords["latitude"] = target["latitude"]

    target_field = xr.DataArray(
        target_values,
        coords=coords.coords,
        dims=data.dims
    )
    return target_field



from typing import Union

import numpy as np
import xarray as xr
from scipy.interpolate import (
    interpn,
    RectBivariateSpline
)

from ._interpolator import (
    BaseInterpolator,
    _create_data_array
)


class ScipyInterpnInterpolator(BaseInterpolator):
    def __init__(
            self,
            method: str,
            **kwargs
    ):
        self.method = method
        self.kwargs = kwargs

    def interpolate_grid(
            self,
            data: xr.DataArray,
            target: xr.DataArray,
    ) -> xr.DataArray:
        latitudes = data.latitude.values
        longitudes = data.longitude.values

        values = data.values

        target_latitudes = target.latitude.values
        target_longitudes = target.longitude.values

        target_x, target_y = np.meshgrid(
            target_longitudes,
            target_latitudes,
        )

        target_values = interpn(
            (latitudes[::-1], longitudes),
            values[::-1, :],
            (target_y, target_x),
            method=self.method,
            **self.kwargs
        )

        target_field =_create_data_array(
            data=data,
            target=target,
            target_values=target_values
        )

        return target_field

    def extract_point(
            self,
            data: xr.DataArray,
            latitude: Union[float, int, list[Union[float, int]]],
            longitude: Union[float, int, list[Union[float, int]]],
    ) -> xr.DataArray:
        latitudes = data.latitude.values
        longitudes = data.longitude.values

        values = data.values

        if isinstance(latitude, list):
            target_latitudes = latitude
        else:
            target_latitudes = [latitude]
        if isinstance(longitude, list):
            target_longitudes = longitude
        else:
            target_longitudes = [longitude]

        target_x, target_y = np.meshgrid(
            target_longitudes,
            target_latitudes,
        )

        target_values = interpn(
            (latitudes[::-1], longitudes),
            values[::-1, :],
            (target_y, target_x),
            method=self.method,
            **self.kwargs
        )

        target_grid = xr.DataArray(
            np.zeros((len(target_latitudes), len(target_longitudes))),  # for xarray v0.10.3
            coords=[
                ("latitude", target_latitudes),
                ("longitude", target_longitudes)
            ],
        )

        target_field = _create_data_array(
            data=data,
            target=target_grid,
            target_values=target_values
        )

        return target_field


class ScipyRectBivariateSplineInterpolator(BaseInterpolator):
    def __init__(
            self,
            method: str,
            **kwargs
    ):
        self.method = method
        self.kwargs = kwargs

    def interpolate_grid(
            self,
            data: xr.DataArray,
            target: xr.DataArray,
    ) -> xr.DataArray:
        latitudes = data.latitude.values
        longitudes = data.longitude.values

        values = data.values

        target_latitudes = target.latitude.values
        target_longitudes = target.longitude.values

        rbs = RectBivariateSpline(
            latitudes[::-1],
            longitudes,
            values[::-1, :],
            **self.kwargs
        )

        if target_latitudes[0] > target_latitudes[-1]:
            target_values = rbs(
                target_latitudes[::-1],
                target_longitudes,
            )[::-1, :]
        else:
            target_values = rbs(
                target_latitudes,
                target_longitudes,
            )


        target_field =_create_data_array(
            data=data,
            target=target,
            target_values=target_values
        )

        return target_field

    def extract_point(
            self,
            data: xr.DataArray,
            latitude,
            longitude,
    ) -> xr.DataArray:
        latitudes = data.latitude.values
        longitudes = data.longitude.values

        values = data.values

        if isinstance(latitude, list):
            target_latitudes = latitude
        else:
            target_latitudes = [latitude]
        if isinstance(longitude, list):
            target_longitudes = longitude
        else:
            target_longitudes = [longitude]

        rbs = RectBivariateSpline(
            latitudes[::-1],
            longitudes,
            values[::-1, :],
            **self.kwargs
        )

        if target_latitudes[0] > target_latitudes[-1]:
            target_values = rbs(
                target_latitudes[::-1],
                target_longitudes,
            )[::-1, :]
        else:
            target_values = rbs(
                target_latitudes,
                target_longitudes,
            )

        target_grid = xr.DataArray(
            np.zeros((len(target_latitudes), len(target_longitudes))),  # for xarray v0.10.3
            coords=[
                ("latitude", target_latitudes),
                ("longitude", target_longitudes)
            ]
        )

        target_field = _create_data_array(
            data=data,
            target=target_grid,
            target_values=target_values
        )

        return target_field



from typing import Optional

from .wgrib2_short_name import WGRIB2_SHORT_NAME_TABLE
from .cemc_param_table import CEMC_PARAM_TABLE


def find_short_name(discipline: int, category: int, number: int) -> Optional[str]:
    """
    Get parameter's short name from WGRIB2 and CEMC param table.

    Parameters
    ----------
    discipline
        GRIB key discipline
    category
        GRIB key parameterCategory
    number
        GRIB key parameterNumber

    Returns
    -------
    Optional[str]
        short name if found, or None if not.
    """
    df = WGRIB2_SHORT_NAME_TABLE.query(
        f'discipline=={discipline} & parameterCategory=={category} & parameterNumber=={number}'
    )
    if not df.empty:
        return df.iloc[0]["short_name"]

    param_df = CEMC_PARAM_TABLE.query(
        f'discipline=={discipline} & category=={category} & number=={number}'
    )
    if not param_df.empty:
        return param_df.iloc[0]["name"]

    return None



from dataclasses import dataclass, asdict
from typing import Dict, List, Union, Optional

from reki.format.grads import load_field_from_file


@dataclass
class QueryOption:
    parameter: str
    level_type: Optional[str]
    level: Optional[Union[int, float, List]]


@dataclass
class TestCase:
    query: QueryOption
    expected_level_type: str
    expected_level: Union[int, float, List]


def test_embedded_name(file_path, modelvar_file_path, forecast_time):
    test_cases = [
        TestCase(query=QueryOption("t", "pl", 250), expected_level_type="pl", expected_level=250),
        TestCase(query=QueryOption("u", "pl", 10.0), expected_level_type="pl", expected_level=10.0)
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert field.coords[test_case.expected_level_type] == test_case.expected_level

    test_cases = [
        TestCase(query=QueryOption("u", "ml", 30), expected_level_type="ml", expected_level=30),
        TestCase(query=QueryOption("w", "ml", 50), expected_level_type="ml", expected_level=50)
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            modelvar_file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert field.coords[test_case.expected_level_type] == test_case.expected_level


def test_index(file_path, forecast_time):
    test_cases = [
        TestCase(query=QueryOption("mslb", "index", 0), expected_level_type="level", expected_level=1000),
        TestCase(query=QueryOption("mslb", "index", 1), expected_level_type="level", expected_level=975),
        TestCase(query=QueryOption("mslb", "index", 2), expected_level_type="level", expected_level=950),
        TestCase(query=QueryOption("mslb", "index", 3), expected_level_type="level", expected_level=925)
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert field.coords[test_case.expected_level_type] == test_case.expected_level


def test_single(file_path, forecast_time):
    test_cases = [
        TestCase(query=QueryOption("glw", "single", None), expected_level_type="level", expected_level=0),
        TestCase(query=QueryOption("ts", "single", None), expected_level_type="level", expected_level=0)
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert field.coords[test_case.expected_level_type] == test_case.expected_level


def test_none(file_path, forecast_time):
    test_cases = [
        TestCase(query=QueryOption("rainc", None, None), expected_level_type="level", expected_level=0),
        TestCase(query=QueryOption("qfx", None, None), expected_level_type="level", expected_level=0)
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert field.coords[test_case.expected_level_type] == test_case.expected_level




from typing import Union, List, Dict
from pathlib import Path
from io import StringIO

import pandas as pd


def load_table_from_file(
        file_path: Union[str, Path],
        sep=r"\s+|,",
        **kwargs
) -> pd.DataFrame:
    """
    Load data from table file using ``pd.read_table``.

    Parameters
    ----------
    file_path
    sep
        ``sep`` option in ``pd.read_table()``
    **kwargs
        additional arguments for ``pd.read_table()``

    Returns
    -------
    pd.DataFrame
    """
    df = pd.read_table(
        file_path,
        sep=sep,
        **kwargs
    )
    return df


NWPC_OBS_CONFIG = {
    "RGWST": {
        "columns": [
            "Station_Id_C", "Station_Id_d", "Lat", "Lon", "Alti", "Day", "Hour",
            "PRS_Sea", "TEM", "DPT", "WIN_D_INST", "WIN_S_INST", "PRE_1h", "PRE_6h", "PRE_24h", "PRS",
            "Q_PRS_Sea", "Q_TEM", "Q_DPT", "Q_WIN_D_INST", "Q_WIN_S_INST", "Q_PRE_1h", "Q_PRE_6h", "Q_PRE_24h", "Q_PRS",
        ]
    },
    "RSURF": {
        "columns": [
            "Station_Id_C", "Station_Id_d", "Lat", "Lon", "Alti", "Day", "Hour", "Min",
            "PRS_HWC", "EVSS", "GPH", "TEM", "DPT", "WIN_D", "WIN_S", "Time_Dev_WQ", "Lat_Dev", "Lon_Dev",
            "Q_PRS_HWC", "Q_GPH", "Q_TEM", "Q_DPT", "Q_WIN_D", "Q_WIN_S"
        ]
    },
    "R2CWE": {
        "columns": [
            "SATE_ID", "Lat", "Lon", "Day", "HourMin", "V02023",
            "PRS_HWC", "V12071", "WIN_D", "WIN_S",
            "Q07004_001", "Q12071_001", "Q11001_001", "Q11002_001",
        ]
    },
    "RSING": {
        "columns": [
            "Station_Id_C", "Lat", "Lon", "Day", "HourMin", "Flight_Stat",
            "Flight_Heigh", "TEM", "WIN_D", "WIN_S",
            "Q_Flight_Heigh", "Q_TEM", "Q_WIN_D", "Q_WIN_S",
        ]
    },
    "RTEMP": {
        "header_columns": [
            "Station_Id_d", "Lat", "Lon", "Alti", "Day", "HourMin", "count"
        ],
        "columns": [
            "PRS_HWC", "EVSS", "GPH", "TEM", "DPT", "WIN_D", "WIN_S", "Time_Dev_WQ", "Lat_Dev", "Lon_Dev",
            "Q_PRS_HWC", "Q_GPH", "Q_TEM", "Q_DPT", "Q_WIN_D", "Q_WIN_S",
        ]
    }
}


def load_nwpc_obs_from_file(
    file_path: Union[Path, str],
) -> pd.DataFrame:
    """
    Load obs text file generated by GRAPES models.

    Supported obs file:

    * RGWST
    * RSURF
    * R2CWE
    * RSING
    * RTEMP

    Parameters
    ----------
    file_path
        obs file path.
    Returns
    -------
    pd.DataFrame
    """
    with open(file_path, "r") as f:
        lines = f.readlines()

    info_line = lines.pop(0)
    tokens = info_line.split()
    if len(tokens) not in (9, 10):
        raise ValueError("first line must have 9 or 10 tokens")
    obs_type = tokens[0]
    obs_time = pd.Timestamp(
        year=int(tokens[1]),
        month=int(tokens[2]),
        day=int(tokens[3]),
        hour=int(tokens[4])
    )

    config = NWPC_OBS_CONFIG.get(obs_type)

    if obs_type == "RTEMP":
        df = _load_nwpc_rtemp_data(lines, config)
    else:
        df = _load_nwpc_obs_table(lines, config)

    df["obs_time"] = obs_time
    return df


def _load_nwpc_obs_table(lines, config):
    sio = StringIO("")
    t = ""
    for index, line in enumerate(lines):
        if index % 2 == 0 and index > 0:
            t = t.strip()
            sio.write(t + "\n")
            t = ""
        t += "     " + line.strip()
    sio.seek(0)

    df = pd.read_table(
        sio,
        sep=r"\s+",
        names=config["columns"]
    )
    sio.close()
    return df


def _load_nwpc_rtemp_data(lines: List[str], config: Dict) -> pd.DataFrame:
    dfs = []
    while len(lines) > 0:
        line = lines.pop(0)
        sio = StringIO(line)
        station_header = pd.read_table(
            sio,
            sep=r"\s+",
            names=config["header_columns"]
        ).loc[0]
        sio.close()
        count = int(station_header["count"])

        sio = StringIO(line)
        for _ in range(count):
            sio.write(lines.pop(0))
        sio.seek(0)
        station_df = pd.read_table(
            sio,
            sep=r"\s+",
            names=config["columns"]
        )
        sio.close()

        # NEED improvement
        for i, v in list(station_header.items())[:-1]:
            station_df[i] = v

        columns_order = config["header_columns"][:-1] + config["columns"]
        station_df = station_df[columns_order]

        dfs.append(station_df)

    df = pd.concat(dfs, ignore_index=True)
    return df



from google.protobuf import json_format

from .transport_pb2 import RawField


def convert_to_json(raw_field):
    return json_format.MessageToJson(raw_field)


def load_from_json(json_string, cls):
    return json_format.Parse(json_string, cls())



import datetime
from typing import Union, Optional

import pandas as pd
import xarray as xr
import requests

from reki.format.grib.eccodes.bytes import create_message_from_bytes
from reki.format.grib.eccodes._xarray import create_data_array_from_message

from .transport import RawField, load_from_json


def load_message(
        system: str,
        stream: str,
        data_type: str,
        data_name: str,
        start_time: Union[datetime.datetime, pd.Timestamp, str],
        forecast_time: Union[pd.Timedelta, str],
        parameter: str,
        server: str,
        level_type: Optional[str] = None,
        level: Optional[int] = None,
        data_class: str = "od",
) -> Optional[int]:
    url = f"http://{server}/api/v1/gdata/fetch/field"

    if not isinstance(start_time, str):
        start_time = start_time.strftime("%Y-%m-%d %H:%M:%S")

    if not isinstance(start_time, str):
        forecast_time = forecast_time.isoformat()

    result = requests.post(url, json={
        "query": {
            "class": data_class,
            "system": system,
            "stream": stream,
            "type": data_type,
            "name": data_name,
            "start_time": start_time,
            "forecast_time": forecast_time,
            "parameter": parameter,
            "level_type": level_type,
            "level": level,
        }
    })

    if result.json()["status"] != "complete":
        return None

    raw_field = result.json()["raw_field"]
    message = load_from_json(raw_field, RawField)

    raw_bytes = message.raw_bytes

    grib_message = create_message_from_bytes(raw_bytes)
    return grib_message


def load_field(
        system: str,
        stream: str,
        data_type: str,
        data_name: str,
        start_time: Union[datetime.datetime, pd.Timestamp, str],
        forecast_time: Union[pd.Timedelta, str],
        parameter: str,
        server: str,
        level_type: str = None,
        level: int = None,
        data_class: str = "od",
) -> Optional[xr.DataArray]:
    message = load_message(
        data_class=data_class,
        system=system,
        stream=stream,
        data_type=data_type,
        data_name=data_name,
        start_time=start_time,
        forecast_time=forecast_time,
        parameter=parameter,
        level_type=level_type,
        level=level,
        server=server,
    )
    if message is None:
        return None

    return create_data_array_from_message(message)



from typing import Optional, Union, Dict, Literal


def fix_level_type(
        level_type: Optional[Union[str, Dict]],
        engine: Literal["eccodes", "cfgrib"] = "eccodes",
) -> Optional[Union[str, Dict]]:
    """
    Convert level type into dict of GRIB key(s).

    Convert some string level to GRIB key(s).

        - pl
        - ml
        - sfc

    If level_type is Dict, or not one of above strings, just return level_type value.

    Notes
    -----
    ``typeOfFirstFixedSurface`` is not working in cfgrib engine.

    Parameters
    ----------
    level_type
    engine
        load engine:

        - cfgrib: use cfgrib to load message
        - eccodes: use eccodes to load message

    """
    if level_type is None:
        return level_type
    if isinstance(level_type, Dict):
        return level_type
    if level_type == "pl":
        if engine == "cfgrib":
            return {
                "typeOfLevel": "isobaricInhPa"
            }
        else:
            return {
                "typeOfFirstFixedSurface": 100,
            }
    elif level_type == "sfc":
        return {
            "typeOfLevel": "sfc"
        }
    elif level_type == "ml":
        return {
            "typeOfFirstFixedSurface": 131,
            # "typeOfSecondFixedSurface": 255,
        }
    return level_type



from typing import Union, Optional

import xarray as xr
import numpy as np


def extract_region(
        data: xr.DataArray,
        start_longitude: Union[float, int],
        end_longitude: Union[float, int],
        start_latitude: Union[float, int],
        end_latitude: Union[float, int],
        longitude_step: Optional[Union[float, int]] = None,
        latitude_step: Optional[Union[float, int]] = None,
) -> xr.DataArray:
    """
    extract region from gridded data array.

    Parameters
    ----------
    data
    start_longitude
    end_longitude
    start_latitude
    end_latitude
    longitude_step
    latitude_step

    Returns
    -------
    xr.DataArray
    """
    if longitude_step is None and latitude_step is None:
        return data.sel(
            longitude=slice(start_longitude, end_longitude),
            latitude=slice(start_latitude, end_latitude)
        )
    elif latitude_step is not None and latitude_step is not None:
        orig_lat_step = data.latitude[1] - data.latitude[0]
        orig_lon_step = data.longitude[1] - data.longitude[0]
        return data.sel(
            latitude=xr.DataArray(np.arange(start_latitude, end_latitude + orig_lat_step/10.0, latitude_step), dims="latitude"),
            longitude=xr.DataArray(np.arange(start_longitude, end_longitude + orig_lon_step/10.0, longitude_step), dims="longitude")
        )
    else:
        raise ValueError("longitude_step and latitude_step must be set together.")



from pathlib import Path
from typing import List, Dict, Union, Optional

import eccodes

from reki.format.grib.common._level import fix_level_type
from reki.format.grib.common._parameter import convert_parameter
from ._util import (
    _check_message,
)


def load_bytes_from_file(
        file_path: Union[str, Path],
        parameter: Union[str, Dict],
        level_type: str = None,
        level: int = None,
) -> Optional[bytes]:
    """
    Load one message from grib file and return message's original bytes.

    Parameters
    ----------
    file_path
    parameter
    level_type
    level

    Returns
    -------
    bytes or None

    Examples
    --------
    Load bytes of 850hPa temperature from GRAPES GFS GMF and create GRIB message using `eccodes.codes_new_from_message`.

    >>> file_path = "/sstorage1/COMMONDATA/OPER/NWPC/GRAPES_GFS_GMF/Prod-grib/2020031721/ORIG/gmf.gra.2020031800105.grb2"
    >>> message_bytes = load_bytes_from_file(
    ...     file_path=file_path,
    ...     parameter="t",
    ...     level_type="pl",
    ...     level=850,
    ... )
    >>> message = eccodes.codes_new_from_message(message_bytes)
    >>> values = eccodes.codes_get_double_array(message, "values")
    >>> print(len(values))
    1036800

    """
    offset = 0

    fixed_level_type = fix_level_type(level_type)

    parameter = convert_parameter(parameter)

    with open(file_path, "rb") as f:
        while True:
            message_id = eccodes.codes_grib_new_from_file(f)
            length = eccodes.codes_get(message_id, "totalLength")
            if message_id is None:
                return None
            if not _check_message(message_id, parameter, fixed_level_type, level):
                eccodes.codes_release(message_id)
                offset += length
                continue
            return eccodes.codes_get_message(message_id)


def create_message_from_bytes(raw_message: bytes) -> Optional[int]:
    """
    Create **the first** message from raw bytes.

    The returned message should be released using ``eccodes.codes_release`` manually.

    Parameters
    ----------
    raw_message

    Returns
    -------
    int or None
    """
    return eccodes.codes_new_from_message(raw_message)


def create_messages_from_bytes(raw_message: bytes) -> List[int]:
    """
    Create all messages from raw bytes.

    The returned messages should be released using ``eccodes.codes_release`` manually.

    Parameters
    ----------
    raw_message

    Returns
    -------
    int or None

    Examples
    --------
    Load raw bytes of temperature at 500hPa and 850hPa and reload them from raw bytes.
    Print level value of all returned messages.

    >>> file_path = "/sstorage1/COMMONDATA/OPER/NWPC/GRAPES_GFS_GMF/Prod-grib/2020031721/ORIG/gmf.gra.2020031800105.grb2"
    >>> message_bytes_t850 = load_bytes_from_file(
    ...     file_path=file_path,
    ...     parameter="t",
    ...     level_type="pl",
    ...     level=850,
    ... )
    >>> message_bytes_t500 = load_bytes_from_file(
    ...     file_path=file_path,
    ...     parameter="t",
    ...     level_type="pl",
    ...     level=500,
    ... )
    >>> message_bytes = message_bytes_t500 + message_bytes_t850
    >>> messages = create_messages_from_bytes(message_bytes)
    >>> for message in messages:
    ...     print(eccodes.codes_get(message, "level"))
    500
    850

    """
    messages = []
    message_bytes = raw_message

    while len(message_bytes) > 0:
        message = create_message_from_bytes(message_bytes)
        if message is None:
            break
        message_size = eccodes.codes_get_message_size(message)
        messages.append(message)
        message_bytes = message_bytes[message_size:]
    return messages



from typing import Union, Literal

import xarray as xr

from ._interpolator import BaseInterpolator


class XarrayInterpolator(BaseInterpolator):
    def __init__(
            self,
            method: Literal["linear", "nearest"],
            **kwargs
    ):
        self.method = method
        self.kwargs = kwargs

    def interpolate_grid(
            self,
            data,
            target
    ) -> xr.DataArray:
        target_field = data.interp_like(
            target,
            method=self.method,
            **self.kwargs
        )

        return target_field

    def extract_point(
            self,
            data: xr.DataArray,
            latitude: Union[int, float],
            longitude: Union[int, float],
    ) -> xr.DataArray:
        value = data.interp(
            latitude=latitude,
            longitude=longitude,
            method=self.method,
        )
        return value



from typing import Optional, Union, List, Dict

from ._util import _check_message
from ._level import _fix_level
from reki.format.grib.common import convert_parameter


def check_message(
        message_id,
        parameter: Optional[Union[str, Dict]],
        level_type: Optional[Union[str, List[str]]],
        level: Optional[Union[int, List[int], Dict]],
        **kwargs,
) -> bool:
    """
    Check whether GRIB message fits conditions.

    Parameters
    ----------
    message_id
    parameter
    level_type
    level
    kwargs

    Returns
    -------
    bool
    """
    # parameter = _convert_parameter(parameter)
    level_type, _ = _fix_level(level_type, None)

    return _check_message(
        message_id,
        parameter=parameter,
        level_type=level_type,
        level=level,
        **kwargs
    )



from typing import Union, Optional, Dict, Literal
from pathlib import Path

import xarray as xr

from .cfgrib import load_fields_from_file
from .common import fix_level_type


def load_field_from_file(
        file_path: Union[str, Path],
        parameter: Union[str, Dict],
        level_type: Optional[Union[str, Dict]] = None,
        level: Optional[int] = None,
        engine: Literal["eccodes", "cfgrib"] = "eccodes",
        **kwargs
) -> Optional[xr.DataArray]:
    """
    Load **one** field from GRIB2 file. Default engine is eccodes.

    Parameters
    ----------
    file_path
    parameter
    level_type
        level type, pl, sfc, ml, or use ecCodes key `typeOfLevel`, or set ecCodes keys directly.
    level
    engine
        GRIB decoding engine, default

        * `eccodes`: use eccodes
        * `cfgrib`: use cfgrib

    kwargs
        other parameters used by engine.

    Returns
    -------
    Optional[xr.DataArray]
        DataArray if found one field, or None if not.
    """
    fixed_level_type = fix_level_type(level_type, engine=engine)
    if engine == "cfgrib":
        from .cfgrib import load_field_from_file
        return load_field_from_file(
            file_path,
            parameter,
            fixed_level_type,
            level,
            **kwargs,
        )
    elif engine == "eccodes":
        from .eccodes import load_field_from_file
        return load_field_from_file(
            file_path,
            parameter,
            fixed_level_type,
            level,
            **kwargs
        )
    else:
        raise ValueError(f"engine {engine} is not supported")



import pandas as pd
from typing import Optional

from .grads_ctl import GradsCtl
from .grads_record_handler import GradsRecordHandler


class GradsDataHandler(object):
    """
    Parse GrADS binary data file with a ctl file.

    Get record from GrADS binary data file.
    """
    def __init__(self, a_grads_ctl: GradsCtl = None):
        if a_grads_ctl is None:
            a_grads_ctl = GradsCtl()

        self.grads_ctl = a_grads_ctl

    def get_offset_by_record_index(self, record_index: int):
        """
        get offset by record index

        Parameters
        ----------
        record_index
            record index

        Returns
        -------
        int
            record offset
        """
        if record_index >= len(self.grads_ctl.record):
            raise ValueError('record_index is too large.')

        offset = 0
        nx = self.grads_ctl.xdef['count']
        ny = self.grads_ctl.ydef['count']
        if 'sequential' in self.grads_ctl.options:
            offset += (nx * ny + 2) * 4 * record_index
        else:
            offset += nx * ny * 4 * record_index

        return offset

    def get_offset_by_index(self, var_index: int, level_index: int = 0):
        """
        get record offset by variable and level index.

        :param var_index:
        :param level_index:
        :return: record offset
        """

        record = self.get_record_by_index(var_index, level_index)
        return record.offset

    def find_record(
            self,
            name: str,
            level: int = 0,
            level_type: str = 'multi',
            valid_time: Optional[pd.Timestamp] = None,
            forecast_time: Optional[pd.Timedelta] = None,
    ):
        """
        find record index by field name, level value, level type.

        Parameters
        ----------
        name
            field name, found name in vars section of ctl files.
        level
            level value
        level_type
            multi or single
        valid_time
        forecast_time

        Returns
        -------
        GradsRecordHandler or None
        """
        cur_i = 0
        if level_type == 'single':
            a_level = 0
        elif level is not None:
            a_level = float(level)
        else:
            a_level = level

        while cur_i < len(self.grads_ctl.record):
            cur_record = self.grads_ctl.record[cur_i]
            if (
                cur_record['name'] == name
                and (level_type is None or cur_record['level_type'] == level_type)
                and (a_level is None or cur_record['level'] == a_level)
                and (valid_time is None or cur_record["valid_time"] == valid_time)
                and (forecast_time is None or cur_record["forecast_time"] == forecast_time)
            ):
                break
            cur_i += 1
        if cur_i < len(self.grads_ctl.record):
            offset = self.get_offset_by_record_index(cur_record["record_index"])
            record = GradsRecordHandler(self.grads_ctl, cur_i, offset)
            return record
        else:
            return None

    def get_record_by_index(self, var_index: int, level_index: int = 0):
        """
        get record from variable and level index.

        Parameters
        ----------
        var_index
        level_index

        Returns
        -------
        GradsRecordHandler
        """
        if var_index >= len(self.grads_ctl.vars):
            raise ValueError("variable index is too large.")

        var_levels = self.grads_ctl.vars[var_index]['levels']
        if 0 < var_levels <= level_index:
            raise ValueError("level index is too large.")

        # calculate record index
        pos = 0
        for a_var_index in range(0, var_index):
            levels = self.grads_ctl.vars[a_var_index]['levels']
            if levels == 0:
                pos += 1
            else:
                pos += levels

        pos += level_index

        offset = self.get_offset_by_record_index(pos)

        record = GradsRecordHandler(self.grads_ctl, pos, offset, var_index=var_index, level_index=level_index)
        return record



from pathlib import Path
from typing import Dict, Union, Optional

import yaml


def find_config(config_dir: Union[str, Path], data_type: str, data_class: str = "od") -> Optional[Path]:
    """
    Find config YAML for some data type by combine ``config_dir``, ``data_type`` and ".yaml".

    Parameters
    ----------
    config_dir
    data_type
    data_class

    Returns
    -------
    Optional[Path]
        config file path, or None if not found.

    Examples
    --------
    Load config file for operation CMA-MESO 3KM grib2 orig product.

    .. code-block:: pycon

        >>> find_config("/path/to/config_dir", "cma_meso_3km/grib2/orig", "od")
        /path/to/config_dir/od/cma_meso_3km/grib2/orig.yaml

    If config YAML file is not found, return None.

    .. code-block:: pycon

        >>> find_config("/path/to/config_dir", "no_system/grib2/orig", "od")
        None

    """
    config_file_path = Path(config_dir, data_class, data_type + ".yaml")
    if config_file_path.is_file():
        return config_file_path
    else:
        return None


def load_config(config_file_path: Union[str, Path]) -> Dict:
    """
    load config dict from config file.

    Parameters
    ----------
    config_file_path

    Returns
    -------
    Dict
        config dict object.
    """
    with open(config_file_path) as config_file:
        config = yaml.safe_load(config_file)
        return config


def get_default_local_config_path() -> Path:
    """
    return default local config path in this package.
    """
    return Path(Path(__file__).parent, "conf").absolute()



from pathlib import Path
from typing import Union, Dict, Optional

import xarray as xr

from reki._util import _load_first_variable


def load_field_from_file(
        file_path: str or Path,
        parameter: Union[str, Dict] = None,
        level_type: Union[str, Dict] = None,
        level: Optional[Union[int, float]] = None,
        **kwargs
) -> Optional[xr.DataArray]:
    """
    Load **one** field from NetCDF file.

    Parameters
    ----------
    file_path
    parameter
    level_type
        level type, default is level.
    level
    **kwargs
        other parameters used by engine.

    Returns
    -------
    DataArray or None:
        DataArray if found one field, or None if not.
    """
    ds = xr.open_dataset(file_path)
    if parameter is None:
        field = _load_first_variable(ds)
    else:
        field = ds[parameter]

    if level is not None:
        if level_type is None:
            level_type = "level"
        field = field.loc[{
            level_type: level
        }]

    return field



from dataclasses import dataclass, asdict
from typing import Union, Dict, Optional, List

import pytest

from reki.format.grib.eccodes import load_field_from_file


@dataclass
class QueryOption:
    parameter: str
    level_type: Union[str, Dict]
    level: Optional[Union[float, Dict]]


@dataclass
class TestCase:
    query: QueryOption


def test_embedded_level_name(file_path, modelvar_file_path):
    """
    embedded level type:

    * pl
    * surface
    * ml
    """
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=850),
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="sfc", level=None),
        ),
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None

    test_cases = [
        TestCase(
            query=QueryOption(parameter="u", level_type="ml", level=10)
        )
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            modelvar_file_path,
            **asdict(test_case.query)
        )
        assert field is not None


def test_type_of_level(file_path):
    test_cases = [
        TestCase(query=QueryOption(parameter="t", level_type="isobaricInhPa", level=850)),
        TestCase(query=QueryOption(parameter="t", level_type="isobaricInPa", level=50)),
        TestCase(query=QueryOption(parameter="acpcp", level_type="surface", level=None)),
        TestCase(query=QueryOption(parameter="tmax", level_type="heightAboveGround", level=2)),
        TestCase(query=QueryOption(parameter="lcc", level_type="entireAtmosphere", level=None)),
        TestCase(query=QueryOption(parameter="ulwrf", level_type="nominalTop", level=None)),
        TestCase(query=QueryOption(parameter="pwat", level_type="atmosphere", level=None)),
        TestCase(query=QueryOption(parameter="prmsl", level_type="meanSea", level=None)),
        TestCase(query=QueryOption(parameter="t", level_type="depthBelowLandLayer", level={"first_level": 0, "second_level": 0.1})),
        TestCase(query=QueryOption(parameter="vwsh", level_type="heightAboveGroundLayer", level=1000))
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None


def test_grib_key(modelvar_file_path):
    test_cases = [
        TestCase(query=QueryOption(parameter="u", level_type={"typeOfFirstFixedSurface": 131}, level=10))
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            modelvar_file_path,
            **asdict(test_case.query)
        )
        assert field is not None



"""
Warnings
--------
This module is only for backward compatibility.

Please use ``reki.format.grib.eccodes`` instead.
"""
from reki.format.grib.eccodes import *

import warnings
warnings.warn("Please use `reki.format.grib.eccodes` instead.", FutureWarning)



from typing import Dict, List, Union, Optional, Tuple


def _fix_level(
        level_type: Optional[Union[str, Dict, List]],
        level_dim: Optional[str],
) -> Tuple[str or dict, str]:
    if level_type is None:
        return level_type, level_dim

    if isinstance(level_type, dict):
        return level_type, level_dim

    if isinstance(level_type, List):
        _, level_dim = _fix_level(level_type[0], level_dim)
        level_type = [_fix_level(cur_level, level_dim)[0] for cur_level in level_type]
        return level_type, level_dim

    if level_type == "pl":
        if level_dim is None:
            level_dim = level_type
        return {
            "typeOfFirstFixedSurface": 100,
        }, level_dim
    elif level_type == "sfc":
        if level_dim is None:
            level_dim = level_type
        return {
            "typeOfLevel": "surface"
        }, level_dim
    elif level_type == "ml":
        if level_dim is None:
            level_dim = level_type
        return {
            "typeOfFirstFixedSurface": 131,
            # "typeOfSecondFixedSurface": 255,
        }, level_dim
    return level_type, level_dim



from dataclasses import dataclass, asdict
from typing import Union, Dict, Optional, List

import pytest
import eccodes
import numpy as np

from reki.format.grib.eccodes import load_message_from_file, load_messages_from_file


@dataclass
class QueryOption:
    parameter: Union[str, dict]
    level_type: Union[str, Dict]
    level: Optional[Union[float, Dict, List]]


@dataclass
class TestCase:
    query: QueryOption
    expected_keys: Optional[Dict] = None


def test_scalar(file_path, modelvar_file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=1.5),
            expected_keys=dict(typeOfLevel="isobaricInhPa", level=1)
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="isobaricInhPa", level=850),
            expected_keys=dict(typeOfLevel="isobaricInhPa", level=850)
        ),
        TestCase(
            query=QueryOption(parameter="tmax", level_type="heightAboveGround", level=2),
            expected_keys=dict(typeOfLevel="heightAboveGround", level=2)
        )
    ]

    for test_case in test_cases:
        message = load_message_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert message is not None
        for key, expected_value in test_case.expected_keys.items():
            assert eccodes.codes_get(message, key) == expected_value

        eccodes.codes_release(message)

    test_cases = [
        TestCase(
            query=QueryOption(parameter="u", level_type={"typeOfFirstFixedSurface": 131}, level=10),
            expected_keys=dict(typeOfFirstFixedSurface=131, level=10)
        )
    ]
    for test_case in test_cases:
        message = load_message_from_file(
            modelvar_file_path,
            **asdict(test_case.query)
        )
        assert message is not None
        for key, expected_value in test_case.expected_keys.items():
            assert eccodes.codes_get(message, key, ktype=int) == expected_value

        eccodes.codes_release(message)


def test_dict(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="vwsh", level_type="heightAboveGroundLayer", level={"first_level": 1000, "second_level": 0}),
            expected_keys=dict(typeOfLevel="heightAboveGroundLayer", level=1000)
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="depthBelowLandLayer", level={"first_level": 0.1, "second_level": 0.4}),
            expected_keys=dict(typeOfLevel="depthBelowLandLayer", level=0)
        )
    ]
    for test_case in test_cases:
        message = load_message_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert message is not None
        for key, expected_value in test_case.expected_keys.items():
            assert eccodes.codes_get(message, key) == expected_value

        eccodes.codes_release(message)


def test_multi_levels(file_path):
    test_cases = [
        TestCase(query=QueryOption(parameter="t", level_type="pl", level=[850, 925, 1000])),
        TestCase(query=QueryOption(parameter="gh", level_type="isobaricInhPa", level=[850, 925, 1000]))
    ]
    for test_case in test_cases:
        messages = load_messages_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert len(messages) == len(test_case.query.level)
        assert np.array_equal(
            np.sort([eccodes.codes_get(message, "level", ktype=int) for message in messages]),
            np.sort(test_case.query.level)
        )

        for message in messages:
            eccodes.codes_release(message)


@pytest.fixture
def pl_levels():
    return [
        1000,
        975,
        950,
        925,
        900,
        850,
        800,
        750,
        700,
        650,
        600,
        550,
        500,
        450,
        400,
        350,
        300,
        275,
        250,
        225,
        200,
        175,
        150,
        125,
        100,
        70,
        50,
        30,
        20,
        10,
        7,
        5,
        4,
        3,
        2,
        1.5,
        1,
        0.5,
        0.2,
        0.1
    ]


def test_all_levels(file_path, pl_levels):
    test_cases = [
        TestCase(query=QueryOption(parameter="t", level_type="pl", level=pl_levels)),
    ]
    for test_case in test_cases:
        messages = load_messages_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert len(messages) == len(test_case.query.level)
        assert np.array_equal(
            np.sort([eccodes.codes_get(message, "level", ktype=int) for message in messages]),
            np.sort([single_level*100 if single_level < 1 else int(single_level) for single_level in test_case.query.level])
        )

        for message in messages:
            eccodes.codes_release(message)


def test_none_level(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=None),
            expected_keys=dict(typeOfLevel="isobaricInhPa", level=1000),
        ),
        TestCase(
            query=QueryOption(parameter="vwsh", level_type="heightAboveGroundLayer", level=None),
            expected_keys=dict(typeOfLevel="heightAboveGroundLayer", level=1000)
        )
    ]
    for test_case in test_cases:
        message = load_message_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert message is not None
        for key, expected_value in test_case.expected_keys.items():
            assert eccodes.codes_get(message, key) == expected_value

        eccodes.codes_release(message)
